@@ -1,14 +1,22 @@
+# vim: ts=4 sw=4 et ai:
 """This module implements the TftpPacketFactory class, which can take a binary
 buffer, and return the appropriate TftpPacket object to represent it, via the
 parse() method."""
 
-from TftpShared import *
-from TftpPacketTypes import *
 
-class TftpPacketFactory(object):
+import logging
+
+from .TftpPacketTypes import *
+from .TftpShared import *
+
+log = logging.getLogger("tftpy.TftpPacketFactory")
+
+
+class TftpPacketFactory:
     """This class generates TftpPacket objects. It is responsible for parsing
     raw buffers off of the wire and returning objects representing them, via
     the parse() method."""
+
     def __init__(self):
         self.classes = {
             1: TftpPacketRRQ,
@@ -16,13 +24,14 @@ class TftpPacketFactory(object):
             3: TftpPacketDAT,
             4: TftpPacketACK,
             5: TftpPacketERR,
-            6: TftpPacketOACK
-            }
+            6: TftpPacketOACK,
+        }
 
     def parse(self, buffer):
         """This method is used to parse an existing datagram into its
         corresponding TftpPacket object. The buffer is the raw bytes off of
         the network."""
+        tftpassert(len(buffer) > 2, 'Invalid packet size')
         log.debug("parsing a %d byte packet" % len(buffer))
         (opcode,) = struct.unpack("!H", buffer[:2])
         log.debug("opcode is %d" % opcode)
@@ -33,8 +42,7 @@ class TftpPacketFactory(object):
     def __create(self, opcode):
         """This method returns the appropriate class object corresponding to
         the passed opcode."""
-        tftpassert(self.classes.has_key(opcode),
-                   "Unsupported opcode: %d" % opcode)
+        tftpassert(opcode in self.classes, "Unsupported opcode: %d" % opcode)
 
         packet = self.classes[opcode]()
 
