@@ -1,5 +1,4 @@
 # vim: ts=4 sw=4 et ai:
-# -*- coding: utf8 -*-
 """This module implements all state handling during uploads and downloads, the
 main interface to which being the TftpState base class.
 
@@ -11,18 +10,22 @@ is complete, at which point it returns None. That is, unless there is a fatal
 error, in which case a TftpException is returned instead."""
 
 
-from .TftpShared import *
-from .TftpPacketTypes import *
-import os
 import logging
+import os
+import random
+import time
 
-log = logging.getLogger('tftpy.TftpStates')
+from .TftpPacketTypes import *
+from .TftpShared import *
+
+log = logging.getLogger("tftpy.TftpStates")
 
 ###############################################################################
 # State classes
 ###############################################################################
 
-class TftpState(object):
+
+class TftpState:
     """The base class for the states."""
 
     def __init__(self, context):
@@ -45,7 +48,7 @@ class TftpState(object):
                 # Set options to OACK options
                 self.context.options = pkt.options
                 for key in self.context.options:
-                    log.info("    %s = %s" % (key, self.context.options[key]))
+                    log.info(f"    {key} = {self.context.options[key]}")
             else:
                 log.error("Failed to negotiate options")
                 raise TftpException("Failed to negotiate options")
@@ -59,21 +62,25 @@ class TftpState(object):
         # FIXME - put this somewhere else?
         accepted_options = {}
         for option in options:
-            if option == 'blksize':
+            if option == "blksize":
                 # Make sure it's valid.
                 if int(options[option]) > MAX_BLKSIZE:
-                    log.info("Client requested blksize greater than %d "
-                             "setting to maximum" % MAX_BLKSIZE)
+                    log.info(
+                        "Client requested blksize greater than %d "
+                        "setting to maximum" % MAX_BLKSIZE
+                    )
                     accepted_options[option] = MAX_BLKSIZE
                 elif int(options[option]) < MIN_BLKSIZE:
-                    log.info("Client requested blksize less than %d "
-                             "setting to minimum" % MIN_BLKSIZE)
+                    log.info(
+                        "Client requested blksize less than %d "
+                        "setting to minimum" % MIN_BLKSIZE
+                    )
                     accepted_options[option] = MIN_BLKSIZE
                 else:
                     accepted_options[option] = options[option]
-            elif option == 'tsize':
+            elif option == "tsize":
                 log.debug("tsize option is set")
-                accepted_options['tsize'] = 0
+                accepted_options["tsize"] = 0
             else:
                 log.info("Dropping unsupported option '%s'" % option)
         log.debug("Returning these accepted options: %s", accepted_options)
@@ -87,7 +94,6 @@ class TftpState(object):
         blocknumber = self.context.next_block
         # Test hook
         if DELAY_BLOCK and DELAY_BLOCK == blocknumber:
-            import time
             log.debug("Deliberately delaying 10 seconds...")
             time.sleep(10)
         dat = None
@@ -95,16 +101,22 @@ class TftpState(object):
         buffer = self.context.fileobj.read(blksize)
         log.debug("Read %d bytes into buffer", len(buffer))
         if len(buffer) < blksize:
-            log.info("Reached EOF on file %s"
-                % self.context.file_to_transfer)
+            log.info("Reached EOF on file %s" % self.context.file_to_transfer)
             finished = True
         dat = TftpPacketDAT()
         dat.data = buffer
         dat.blocknumber = blocknumber
         self.context.metrics.bytes += len(dat.data)
-        log.debug("Sending DAT packet %d", dat.blocknumber)
-        self.context.sock.sendto(dat.encode().buffer,
-                                 (self.context.host, self.context.tidport))
+        # Testing hook
+        if NETWORK_UNRELIABILITY > 0 and random.randrange(
+                NETWORK_UNRELIABILITY) == 0:
+            log.warning("Skipping DAT packet %d for testing", dat.blocknumber)
+        else:
+            log.debug("Sending DAT packet %d", dat.blocknumber)
+            self.context.sock.sendto(
+                dat.encode().buffer, (self.context.host, self.context.tidport)
+            )
+            self.context.metrics.last_dat_time = time.time()
         if self.context.packethook:
             self.context.packethook(dat)
         self.context.last_pkt = dat
@@ -120,9 +132,15 @@ class TftpState(object):
         log.info("Sending ack to block %d" % blocknumber)
         ackpkt = TftpPacketACK()
         ackpkt.blocknumber = blocknumber
-        self.context.sock.sendto(ackpkt.encode().buffer,
-                                 (self.context.host,
-                                  self.context.tidport))
+        # Testing hook
+        if NETWORK_UNRELIABILITY > 0 and random.randrange(
+                NETWORK_UNRELIABILITY) == 0:
+            log.warning("Skipping ACK packet %d for testing",
+                        ackpkt.blocknumber)
+        else:
+            self.context.sock.sendto(
+                ackpkt.encode().buffer, (self.context.host, self.context.tidport)
+            )
         self.context.last_pkt = ackpkt
 
     def sendError(self, errorcode):
@@ -131,12 +149,12 @@ class TftpState(object):
         log.debug("In sendError, being asked to send error %d", errorcode)
         errpkt = TftpPacketERR()
         errpkt.errorcode = errorcode
-        if self.context.tidport == None:
+        if self.context.tidport is None:
             log.debug("Error packet received outside session. Discarding")
         else:
-            self.context.sock.sendto(errpkt.encode().buffer,
-                                     (self.context.host,
-                                      self.context.tidport))
+            self.context.sock.sendto(
+                errpkt.encode().buffer, (self.context.host, self.context.tidport)
+            )
         self.context.last_pkt = errpkt
 
     def sendOACK(self):
@@ -145,15 +163,16 @@ class TftpState(object):
         log.debug("In sendOACK with options %s", self.context.options)
         pkt = TftpPacketOACK()
         pkt.options = self.context.options
-        self.context.sock.sendto(pkt.encode().buffer,
-                                 (self.context.host,
-                                  self.context.tidport))
+        self.context.sock.sendto(
+            pkt.encode().buffer, (self.context.host, self.context.tidport)
+        )
         self.context.last_pkt = pkt
 
     def resendLast(self):
-        "Resend the last sent packet due to a timeout."
-        log.warning("Resending packet %s on sessions %s"
-            % (self.context.last_pkt, self))
+        """Resend the last sent packet due to a timeout."""
+        assert (self.context.last_pkt is not None)
+        log.warning(
+            f"Resending packet {self.context.last_pkt} on sessions {self}")
         self.context.metrics.resent_bytes += len(self.context.last_pkt.buffer)
         self.context.metrics.add_dup(self.context.last_pkt)
         sendto_port = self.context.tidport
@@ -162,8 +181,9 @@ class TftpState(object):
             # started talking to us yet. That's not good. Maybe it's not
             # there.
             sendto_port = self.context.port
-        self.context.sock.sendto(self.context.last_pkt.encode().buffer,
-                                 (self.context.host, sendto_port))
+        self.context.sock.sendto(
+            self.context.last_pkt.encode().buffer, (self.context.host, sendto_port)
+        )
         if self.context.packethook:
             self.context.packethook(self.context.last_pkt)
 
@@ -198,14 +218,17 @@ class TftpState(object):
 
         else:
             # FIXME: should we be more tolerant and just discard instead?
-            msg = "Whoa! Received future block %d but expected %d" \
-                % (pkt.blocknumber, self.context.next_block)
+            msg = "Whoa! Received future block %d but expected %d" % (
+                pkt.blocknumber,
+                self.context.next_block,
+            )
             log.error(msg)
             raise TftpException(msg)
 
         # Default is to ack
         return TftpStateExpectDAT(self.context)
 
+
 class TftpServerState(TftpState):
     """The base class for server states."""
 
@@ -229,7 +252,7 @@ class TftpServerState(TftpState):
             log.info("Setting tidport to %s" % rport)
 
         log.debug("Setting default options, blksize")
-        self.context.options = { 'blksize': DEF_BLKSIZE }
+        self.context.options = {"blksize": DEF_BLKSIZE}
 
         if options:
             log.debug("Options requested: %s", options)
@@ -238,20 +261,19 @@ class TftpServerState(TftpState):
             sendoack = True
 
         # FIXME - only octet mode is supported at this time.
-        if pkt.mode != 'octet':
-            #self.sendError(TftpErrors.IllegalTftpOp)
-            #raise TftpException("Only octet transfers are supported at this time.")
-            log.warning("Received non-octet mode request. I'll reply with binary data.")
+        if pkt.mode != "octet":
+            # self.sendError(TftpErrors.IllegalTftpOp)
+            # raise TftpException("Only octet transfers are supported at this time.")
+            log.warning(
+                "Received non-octet mode request. I'll reply with binary data.")
 
         # test host/port of client end
         if self.context.host != raddress or self.context.port != rport:
             self.sendError(TftpErrors.UnknownTID)
-            log.error("Expected traffic from %s:%s but received it "
-                            "from %s:%s instead."
-                            % (self.context.host,
-                               self.context.port,
-                               raddress,
-                               rport))
+            log.error(
+                "Expected traffic from %s:%s but received it from %s:%s instead."
+                % (self.context.host, self.context.port, raddress, rport)
+            )
             # FIXME: increment an error count?
             # Return same state, we're still waiting for valid traffic.
             return self
@@ -272,14 +294,16 @@ class TftpServerState(TftpState):
         if pkt.filename.startswith(self.context.root):
             full_path = pkt.filename
         else:
-            full_path = os.path.join(self.context.root, pkt.filename.lstrip('/'))
+            full_path = os.path.join(
+                self.context.root, pkt.filename.lstrip("/"))
 
         # Use abspath to eliminate any remaining relative elements
         # (e.g. '..') and ensure that is still within the server's
         # root directory
         self.full_path = os.path.abspath(full_path)
         log.debug("full_path is %s", full_path)
-        if self.full_path.startswith(os.path.normpath(self.context.root) + "/"):
+        if self.full_path.startswith(
+                os.path.normpath(self.context.root) + os.sep):
             log.info("requested file is in the server root - good")
         else:
             log.warning("requested file is not within the server root - bad")
@@ -294,8 +318,9 @@ class TftpServerState(TftpState):
 class TftpStateServerRecvRRQ(TftpServerState):
     """This class represents the state of the TFTP server when it has just
     received an RRQ packet."""
+
     def handle(self, pkt, raddress, rport):
-        "Handle an initial RRQ packet as a server."
+        """Handle an initial RRQ packet as a server."""
         log.debug("In TftpStateServerRecvRRQ.handle")
         sendoack = self.serverInitial(pkt, raddress, rport)
         path = self.full_path
@@ -306,28 +331,29 @@ class TftpStateServerRecvRRQ(TftpServerState):
             self.context.fileobj = open(path, "rb")
         elif self.context.dyn_file_func:
             log.debug("No such file %s but using dyn_file_func", path)
-            self.context.fileobj = \
-                self.context.dyn_file_func(self.context.file_to_transfer, raddress=raddress, rport=rport)
+            self.context.fileobj = self.context.dyn_file_func(
+                self.context.file_to_transfer, raddress=raddress, rport=rport
+            )
 
             if self.context.fileobj is None:
-                log.debug("dyn_file_func returned 'None', treating as "
-                          "FileNotFound")
+                log.debug(
+                    "dyn_file_func returned 'None', treating as " "FileNotFound")
                 self.sendError(TftpErrors.FileNotFound)
                 raise TftpException("File not found: %s" % path)
         else:
             log.warning("File not found: %s", path)
             self.sendError(TftpErrors.FileNotFound)
-            raise TftpException("File not found: {}".format(path))
+            raise TftpException(f"File not found: {path}")
 
         # Options negotiation.
-        if sendoack and 'tsize' in self.context.options:
+        if sendoack and "tsize" in self.context.options:
             # getting the file size for the tsize option. As we handle
             # file-like objects and not only real files, we use this seeking
             # method instead of asking the OS
             self.context.fileobj.seek(0, os.SEEK_END)
             tsize = str(self.context.fileobj.tell())
             self.context.fileobj.seek(0, 0)
-            self.context.options['tsize'] = tsize
+            self.context.options["tsize"] = tsize
 
         if sendoack:
             # Note, next_block is 0 here since that's the proper
@@ -346,9 +372,11 @@ class TftpStateServerRecvRRQ(TftpServerState):
         # Note, we don't have to check any other states in this method, that's
         # up to the caller.
 
+
 class TftpStateServerRecvWRQ(TftpServerState):
     """This class represents the state of the TFTP server when it has just
     received a WRQ packet."""
+
     def make_subdirs(self):
         """The purpose of this method is to, if necessary, create all of the
         subdirectories leading up to the file to the written."""
@@ -369,7 +397,7 @@ class TftpStateServerRecvWRQ(TftpServerState):
                     os.mkdir(current, 0o700)
 
     def handle(self, pkt, raddress, rport):
-        "Handle an initial WRQ packet as a server."
+        """Handle an initial WRQ packet as a server."""
         log.debug("In TftpStateServerRecvWRQ.handle")
         sendoack = self.serverInitial(pkt, raddress, rport)
         path = self.full_path
@@ -384,8 +412,10 @@ class TftpStateServerRecvWRQ(TftpServerState):
             log.info("Opening file %s for writing" % path)
             if os.path.exists(path):
                 # FIXME: correct behavior?
-                log.warning("File %s exists already, overwriting..." % (
-                    self.context.file_to_transfer))
+                log.warning(
+                    "File %s exists already, overwriting..."
+                    % (self.context.file_to_transfer)
+                )
             # FIXME: I think we should upload to a temp file and not overwrite
             # the existing file until the file is successfully uploaded.
             self.make_subdirs()
@@ -409,34 +439,37 @@ class TftpStateServerRecvWRQ(TftpServerState):
         # Note, we don't have to check any other states in this method, that's
         # up to the caller.
 
+
 class TftpStateServerStart(TftpState):
     """The start state for the server. This is a transitory state since at
     this point we don't know if we're handling an upload or a download. We
     will commit to one of them once we interpret the initial packet."""
+
     def handle(self, pkt, raddress, rport):
         """Handle a packet we just received."""
         log.debug("In TftpStateServerStart.handle")
         if isinstance(pkt, TftpPacketRRQ):
             log.debug("Handling an RRQ packet")
-            return TftpStateServerRecvRRQ(self.context).handle(pkt,
-                                                               raddress,
-                                                               rport)
+            return TftpStateServerRecvRRQ(
+                self.context).handle(pkt, raddress, rport)
         elif isinstance(pkt, TftpPacketWRQ):
             log.debug("Handling a WRQ packet")
-            return TftpStateServerRecvWRQ(self.context).handle(pkt,
-                                                               raddress,
-                                                               rport)
+            return TftpStateServerRecvWRQ(
+                self.context).handle(pkt, raddress, rport)
         else:
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException("Invalid packet to begin up/download: %s" % pkt)
+            raise TftpException(
+                "Invalid packet to begin up/download: %s" % pkt)
+
 
 class TftpStateExpectACK(TftpState):
     """This class represents the state of the transfer when a DAT was just
     sent, and we are waiting for an ACK from the server. This class is the
     same one used by the client during the upload, and the server during the
     download."""
+
     def handle(self, pkt, raddress, rport):
-        "Handle a packet, hopefully an ACK since we just sent a DAT."
+        """Handle a packet, hopefully an ACK since we just sent a DAT."""
         if isinstance(pkt, TftpPacketACK):
             log.debug("Received ACK for packet %d" % pkt.blocknumber)
             # Is this an ack to the one we just sent?
@@ -448,17 +481,23 @@ class TftpStateExpectACK(TftpState):
                     log.debug("Good ACK, sending next DAT")
                     self.context.next_block += 1
                     log.debug("Incremented next_block to %d",
-                        self.context.next_block)
+                              self.context.next_block)
                     self.context.pending_complete = self.sendDAT()
 
             elif pkt.blocknumber < self.context.next_block:
-                log.warning("Received duplicate ACK for block %d"
-                    % pkt.blocknumber)
+                log.warning("Received duplicate ACK for block %d" %
+                            pkt.blocknumber)
                 self.context.metrics.add_dup(pkt)
+                if self.context.metrics.last_dat_time > 0:
+                    if time.time() - self.context.metrics.last_dat_time > self.context.timeout:
+                        raise TftpTimeoutExpectACK(
+                            "Timeout waiting for ACK for block %d" % self.context.next_block)
 
             else:
-                log.warning("Oooh, time warp. Received ACK to packet we "
-                         "didn't send yet. Discarding.")
+                log.warning(
+                    "Oooh, time warp. Received ACK to packet we "
+                    "didn't send yet. Discarding."
+                )
                 self.context.metrics.errors += 1
             return self
         elif isinstance(pkt, TftpPacketERR):
@@ -468,8 +507,10 @@ class TftpStateExpectACK(TftpState):
             log.warning("Discarding unsupported packet: %s" % str(pkt))
             return self
 
+
 class TftpStateExpectDAT(TftpState):
     """Just sent an ACK packet. Waiting for DAT."""
+
     def handle(self, pkt, raddress, rport):
         """Handle the packet in response to an ACK, which should be a DAT."""
         if isinstance(pkt, TftpPacketDAT):
@@ -491,10 +532,13 @@ class TftpStateExpectDAT(TftpState):
 
         else:
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException("Received unknown packet type from peer: " + str(pkt))
+            raise TftpException(
+                "Received unknown packet type from peer: " + str(pkt))
+
 
 class TftpStateSentWRQ(TftpState):
     """Just sent an WRQ packet for an upload."""
+
     def handle(self, pkt, raddress, rport):
         """Handle a packet we just received."""
         if not self.context.tidport:
@@ -533,7 +577,6 @@ class TftpStateSentWRQ(TftpState):
                 return self
 
         elif isinstance(pkt, TftpPacketERR):
-            self.sendError(TftpErrors.IllegalTftpOp)
             raise TftpException("Received ERR from server: %s" % pkt)
 
         elif isinstance(pkt, TftpPacketRRQ):
@@ -546,13 +589,16 @@ class TftpStateSentWRQ(TftpState):
 
         else:
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException("Received unknown packet type from server: %s" % pkt)
+            raise TftpException(
+                "Received unknown packet type from server: %s" % pkt)
 
         # By default, no state change.
         return self
 
+
 class TftpStateSentRRQ(TftpState):
     """Just sent an RRQ packet."""
+
     def handle(self, pkt, raddress, rport):
         """Handle the packet in response to an RRQ to the server."""
         if not self.context.tidport:
@@ -582,7 +628,7 @@ class TftpStateSentRRQ(TftpState):
             log.info("Received DAT from server")
             if self.context.options:
                 log.info("Server ignored options, falling back to defaults")
-                self.context.options = { 'blksize': DEF_BLKSIZE }
+                self.context.options = {"blksize": DEF_BLKSIZE}
             return self.handleDat(pkt)
 
         # Every other packet type is a problem.
@@ -596,16 +642,16 @@ class TftpStateSentRRQ(TftpState):
             raise TftpException("Received WRQ from server while in download")
 
         elif isinstance(pkt, TftpPacketERR):
-            self.sendError(TftpErrors.IllegalTftpOp)
             log.debug("Received ERR packet: %s", pkt)
             if pkt.errorcode == TftpErrors.FileNotFound:
                 raise TftpFileNotFoundError("File not found")
             else:
-                raise TftpException("Received ERR from server: {}".format(pkt))
+                raise TftpException(f"Received ERR from server: {pkt}")
 
         else:
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException("Received unknown packet type from server: %s" % pkt)
+            raise TftpException(
+                "Received unknown packet type from server: %s" % pkt)
 
         # By default, no state change.
         return self
